"""Public interface to learning module."""

import asyncio
from typing import Dict, List, Optional
from functools import partial
from copy import deepcopy
import aiohttp
import tqdm  # type: ignore

from redbrick.common.context import RBContext
from redbrick.utils.pagination import PaginationIterator
from redbrick.utils.rb_label_utils import clean_rb_label, flat_rb_format
from redbrick.utils.async_utils import gather_with_concurrency


class Learning:
    """Perform active learning and upload the results with the RedBrick API."""

    def __init__(
        self, context: RBContext, org_id: str, project_id: str, stage_name: str
    ) -> None:
        """Construct Learning module."""
        self.context = context
        self.org_id = org_id
        self.project_id = project_id
        self.stage_name = stage_name

    def get_learning_info(self) -> Dict:
        """
        Get a dictionary with lightly parsed redbrick response.

        The dictonary has fields:
            labeled
            unlabeled
            taxonomy
            cycle
        """
        cycle = self.context.learning.check_for_job(
            self.org_id, self.project_id, self.stage_name
        )
        if not cycle:
            raise Exception("No available jobs")

        my_iter = PaginationIterator(
            partial(
                self.context.learning.get_batch_of_tasks,
                self.org_id,
                self.project_id,
                self.stage_name,
                100,
            )
        )
        taxonomy, td_type = self.context.learning.get_taxonomy_and_type(
            self.org_id, self.project_id, self.stage_name
        )

        def _parse_labeled_entry(item: Dict) -> Dict:
            items_presigned = item["datapoint"]["itemsPresigned"]
            name = item["datapoint"]["name"]
            items = item["datapoint"]["items"]
            dp_id = item["datapoint"]["dpId"]
            task_id = item["taskId"]
            labels = [clean_rb_label(label) for label in item["groundTruth"]["labels"]]

            return {
                "dpId": dp_id,
                "taskId": task_id,
                "items": items,
                "itemsPresigned": items_presigned,
                "name": name,
                "labels": labels,
            }

        def _parse_unlabeled_entry(item: Dict) -> Dict:
            items_presigned = item["datapoint"]["itemsPresigned"]
            name = item["datapoint"]["name"]
            items = item["datapoint"]["items"]
            dp_id = item["datapoint"]["dpId"]
            task_id = item["taskId"]
            return {
                "taskId": task_id,
                "dpId": dp_id,
                "items": items,
                "itemsPresigned": items_presigned,
                "name": name,
            }

        labeled: List[Dict] = []
        unlabeled: List[Dict] = []

        print("Downloading tasks for active learning")
        for entry in tqdm.tqdm(my_iter, unit=" tasks"):
            # print(entry)
            if entry.get("groundTruth"):
                labeled.append(_parse_labeled_entry(entry))
            else:
                unlabeled.append(_parse_unlabeled_entry(entry))

        return {
            "labeled": labeled,
            "unlabeled": unlabeled,
            "taxonomy": taxonomy,
            "cycle": cycle,
            "type": td_type,
        }

    async def _update_task(
        self, session: aiohttp.ClientSession, cycle: int, task: Dict
    ) -> Optional[Dict]:
        """Attempt to update task."""
        try:
            await self.context.learning.send_batch_learning_results_async(
                session, self.org_id, self.project_id, self.stage_name, cycle, [task]
            )
        except Exception as error:
            print(error)
            point_error = deepcopy(task)
            point_error["error"] = error
            return point_error
        return None

    async def _update_tasks(self, cycle: int, tasks: List[Dict]) -> List[Dict]:
        async with aiohttp.ClientSession() as session:
            coros = [self._update_task(session, cycle, task) for task in tasks]
            temp = await gather_with_concurrency(
                100, coros, "Updating tasks with priorities"
            )
            failed = []
            for val in temp:
                if val:
                    failed.append(val)
            return failed

    def update_tasks(self, cycle: int, tasks: List[Dict]) -> List[Dict]:
        """
        Update tasks with new score and labels.

        Return any tasks that experienced issues.
        """
        temp = asyncio.run(self._update_tasks(cycle, tasks))

        # update cycle
        self.context.learning.set_cycle_status(
            self.org_id, self.project_id, self.stage_name, cycle, "DONE"
        )
        return temp



"""A utility iterator to handle default RedBrick pagination behavior."""

from typing import Any, Dict, Iterable, List, Optional, Callable, Tuple


class PaginationIterator:
    """Construct Labelset Iterator."""

    def __init__(
        self, func: Callable[[Optional[str]], Tuple[List[Dict], Optional[str]]]
    ) -> None:
        """Construct LabelsetIterator."""
        self.cursor: Optional[str] = None
        self.datapointsBatch: Optional[List[Dict]] = None
        self.datapointsBatchIndex: Optional[int] = None

        self.func = func

        self.total = 0

    def __iter__(self) -> Any:
        return self

    def __len__(self) -> int:
        return self.total

    def __next__(self) -> Dict:
        """Get next batch of labels / datapoint."""
        # If cursor is None and current datapointsBatch has been processed
        if (
            self.datapointsBatchIndex is not None
            and self.cursor is None
            and self.datapointsBatch
            and len(self.datapointsBatch) == self.datapointsBatchIndex
        ):
            raise StopIteration

        # If current datapointsBatch is None or we have finished processing current datapointsBatch
        if (
            self.datapointsBatch is None
            or len(self.datapointsBatch) == self.datapointsBatchIndex
        ):

            self.datapointsBatch, self.cursor = self.func(self.cursor)
            self.datapointsBatchIndex = 0

            self.total += len(self.datapointsBatch)

        # Current entry to return
        if self.datapointsBatch and self.datapointsBatchIndex is not None:
            entry = self.datapointsBatch[self.datapointsBatchIndex]
            self.datapointsBatchIndex += 1

            return entry

        raise StopIteration



"""Partial queries to prevent duplication."""
LABEL_SHARD = """
category
attributes {
    name
    value
}
labelid
frameindex
trackid
keyframe
taskclassify
tasklevelclassify
frameclassify
end
bbox2d {
    xnorm
    ynorm
    wnorm
    hnorm
}
point {
    xnorm
    ynorm
}
polyline {
    xnorm
    ynorm
}
polygon {
    xnorm
    ynorm
}
pixel {
    imagesize
    regions
    holes
}
ellipse {
    xcenternorm
    ycenternorm
    xnorm
    ynorm
    rot
}
"""



"""Graphql Client responsible for make API requests."""


from typing import Dict, Any
import requests
import aiohttp


class RBClient:
    """Client to communicate with RedBrick AI GraphQL Server."""

    def __init__(self, api_key: str, url: str, retry_count: int = 5,) -> None:
        """Construct RBClient."""
        self.api_key = api_key
        self.url = url.rstrip("/") + "/graphql/"

    def execute_query(self, query: str, variables: Dict[str, Any]) -> Any:
        """Execute a graphql query."""
        headers = {"ApiKey": self.api_key}

        try:
            response = requests.post(
                self.url, headers=headers, json={"query": query, "variables": variables}
            )
            res = {}

            if "errors" in response.json():
                raise ValueError(response.json()["errors"][0]["message"])
            elif "data" in response.json():
                res = response.json()["data"]
            else:
                res = response.json()
            return res
        except ValueError:
            # print(response.content)
            # print(response.status_code)
            raise

    async def execute_query_async(
        self, aio_client: aiohttp.ClientSession, query: str, variables: Dict[str, Any]
    ) -> Any:
        """Execute a graphql query using asyncio."""
        headers = {"ApiKey": self.api_key}

        try:
            async with aio_client.post(
                self.url, headers=headers, json={"query": query, "variables": variables}
            ) as response:

                response_data = await response.json()
                res = {}
                if "errors" in response_data:
                    raise ValueError(response_data["errors"][0]["message"])
                elif "data" in response_data:
                    res = response_data["data"]
                else:
                    res = response_data
                return res
        except ValueError:
            # print(response.content)
            # print(response.status_code)
            raise



"""Public API to exporting."""

import tqdm  # type: ignore
from typing import List, Dict, Callable, Optional, Tuple
from functools import partial
from redbrick.common.context import RBContext
from redbrick.utils.pagination import PaginationIterator
from redbrick.utils.rb_label_utils import clean_rb_label, flat_rb_format
from redbrick.coco.coco_main import coco_converter


class Export:
    """Primary interface to handling export from a project."""

    def __init__(self, context: RBContext, org_id: str, project_id: str) -> None:
        """Construct Export object."""
        self.context = context
        self.org_id = org_id
        self.project_id = project_id

    def _get_raw_data_ground_truth(self, concurrency: int) -> Tuple[List[Dict], Dict]:
        temp = self.context.export.get_datapoints_output

        my_iter = PaginationIterator(
            partial(temp, self.org_id, self.project_id, concurrency, True)
        )

        general_info = self.context.export.get_output_info(self.org_id, self.project_id)

        def _parse_entry(item: Dict) -> Dict:
            items_presigned = item["itemsPresigned"]
            items = item["items"]
            name = item["name"]
            dp_id = item["dpId"]
            created_by = item["labelData"]["createdByEmail"]
            labels = [clean_rb_label(label) for label in item["labelData"]["labels"]]

            return flat_rb_format(
                labels, items, items_presigned, name, dp_id, created_by
            )

        print("Downloading datapoints")
        return (
            [
                _parse_entry(val)
                for val in tqdm.tqdm(
                    my_iter, unit=" datapoints", total=general_info["datapointCount"]
                )
            ],
            general_info["taxonomy"],
        )

    def _get_raw_data_latest(self, concurrency: int) -> Tuple[List[Dict], Dict]:
        temp = self.context.export.get_datapoints_latest

        my_iter = PaginationIterator(
            partial(temp, self.org_id, self.project_id, concurrency, True)
        )

        general_info = self.context.export.get_output_info(self.org_id, self.project_id)
        datapoint_count = self.context.export.datapoints_in_project(
            self.org_id, self.project_id
        )

        def _parse_entry(item: Dict) -> Dict:
            history_obj = item["history"][0]
            datapoint = history_obj["taskData"]["dataPoint"]
            items_presigned = datapoint["itemsPresigned"]
            items = datapoint["items"]
            name = datapoint["name"]
            dp_id = datapoint["dpId"]
            created_by = history_obj["taskData"]["createdByEmail"]
            labels = [
                clean_rb_label(label) for label in history_obj["taskData"]["labels"]
            ]

            return flat_rb_format(
                labels, items, items_presigned, name, dp_id, created_by
            )

        print("Downloading datapoints")
        return (
            [
                _parse_entry(val)
                for val in tqdm.tqdm(my_iter, unit=" datapoints", total=datapoint_count)
            ],
            general_info["taxonomy"],
        )

    def redbrick_format(
        self, only_ground_truth: bool = True, concurrency: int = 10
    ) -> List[Dict]:
        """Export data into redbrick format."""
        if only_ground_truth:
            datapoints, taxonomy = self._get_raw_data_ground_truth(concurrency)
        else:
            datapoints, taxonomy = self._get_raw_data_latest(concurrency)

        return datapoints

    def coco_format(
        self, only_ground_truth: bool = True, concurrency: int = 10
    ) -> Dict:
        """Export project into coco format."""
        if only_ground_truth:
            datapoints, taxonomy = self._get_raw_data_ground_truth(concurrency)
        else:
            datapoints, taxonomy = self._get_raw_data_latest(concurrency)
        return coco_converter(datapoints, taxonomy)



"""Template creator for the most basic projects."""

from typing import List, Dict, Tuple

OUTPUT_NAME = "Output"
LABEL_NAME = "Label"
MIN_REVIEW = 0
MAX_REVIEW = 6


def _get_middle_stages(reviews: int, passed_name: str) -> Tuple[List[Dict], str]:
    """Get label and review stages."""
    label_stage = {
        "brickName": "manual-labeling",
        "stageName": LABEL_NAME,
        "routing": {"nextStageName": "Review_1" if reviews > 0 else passed_name,},
        "stageConfig": {},
    }
    feedback_stage = {
        "brickName": "feedback",
        "stageName": "Failed_Review",
        "routing": {"feedbackStageName": label_stage["stageName"],},
        "stageConfig": {},
    }

    stages = [label_stage]
    for ii in range(1, reviews + 1):
        stages.append(
            {
                "brickName": "expert-review",
                "stageName": f"Review_{ii}",
                "stageConfig": {},
                "routing": {
                    "passed": passed_name if ii == reviews else f"Review_{ii+1}",
                    "failed": feedback_stage["stageName"],
                },
            }
        )

    if reviews > 0:
        stages.append(feedback_stage)

    return stages, LABEL_NAME


def get_basic_project(reviews: int = 0) -> List[Dict]:
    """Get basic project config with reviews."""
    if reviews < MIN_REVIEW:
        reviews = MIN_REVIEW
    if reviews > MAX_REVIEW:
        reviews = MAX_REVIEW

    middle_stages, entry_point = _get_middle_stages(reviews, OUTPUT_NAME)

    input_stage = {
        "brickName": "labelset-input",
        "routing": {"nextStageName": entry_point,},
        "stageName": "Input",
        "stageConfig": {},
    }
    output_stage = {
        "brickName": "labelset-output",
        "stageName": OUTPUT_NAME,
        "routing": {"nextStageName": "END",},
        "stageConfig": {},
    }

    temp = [input_stage] + middle_stages + [output_stage]

    return temp



from typing import Optional, List, Dict, Tuple
from redbrick.common.export import ExportControllerInterface
from redbrick.common.client import RBClient
from .shards import LABEL_SHARD


class ExportRepo(ExportControllerInterface):
    """Handle API requests to get export data."""

    def __init__(self, client: RBClient) -> None:
        """Construct ExportRepo."""
        self.client = client

    def get_datapoints_input(
        self,
        org_id: str,
        project_id: str,
        first: int = 50,
        presign: bool = False,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get datapoints that were uploaded to the project."""
        raise NotImplementedError()

    def get_output_info(self, org_id: str, project_id: str) -> Dict:
        """Get info about the output labelset and taxonomy."""
        query_string = """
        query ($orgId: UUID!, $name: String!){
            customGroup(orgId: $orgId, name:$name){
                dataType
                taskType
                datapointCount
                taxonomy {
                    name
                    version
                    categories {
                        name
                        children {
                            name
                            classId
                            children {
                                name
                                classId
                                children {
                                    name
                                    classId
                                }
                            }
                        }
                    }
                }
            }

        }
        """

        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "name": project_id + "-output",
        }

        result = self.client.execute_query(query_string, query_variables)

        temp: Dict = result["customGroup"]
        return temp

    def get_datapoints_output(
        self,
        org_id: str,
        project_id: str,
        first: int = 50,
        presign: bool = False,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get datapoints that have made it to the output of the project."""
        query_string = (
            """
        query ($orgId: UUID!, $name: String!,$first: Int, $cursor: String){
        customGroup(orgId: $orgId, name:$name){
            datapointsPaged(first:$first, after:$cursor) {
            entries {
                dpId
                name
                itemsPresigned:items (presigned:true)
                items(presigned:false)
                labelData(customGroupName: $name){
                createdByEmail
                labels(interpolate: true) {
                    %s
                    }
                }


            }
            cursor
            }
        }
        }

        """
            % LABEL_SHARD
        )
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "name": project_id + "-output",
            "cursor": cursor,
            "first": first,
        }

        result = self.client.execute_query(query_string, query_variables)

        return (
            result["customGroup"]["datapointsPaged"]["entries"],
            result["customGroup"]["datapointsPaged"]["cursor"],
        )

    def datapoints_in_project(self, org_id: str, project_id: str) -> int:
        """Get number of datapoints in project."""
        query_string = """
        query($orgId: UUID!, $projectId: UUID!) {
            tasksPaged(
                orgId: $orgId
                projectId: $projectId
            ) {
                count
            }
        }
        """
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
        }

        result = self.client.execute_query(query_string, query_variables)

        return int(result["tasksPaged"]["count"])

    def get_datapoints_latest(
        self,
        org_id: str,
        project_id: str,
        first: int = 50,
        presign: bool = False,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get the latest datapoints."""
        query_string = (
            """
        query($orgId: UUID!, $projectId: UUID!, $first: Int, $cursor: String) {
            tasksPaged(
                orgId: $orgId
                projectId: $projectId
                first: $first
                after: $cursor
            ) {
                entries {
                history(latest: true) {
                    taskData {
                    dataPoint {
                        dpId
                        name
                        itemsPresigned: items(presigned: true)
                        items(presigned: false)
                    }
                    createdByEmail
                    labels(interpolate: true) {
                        %s
                    }
                    }
                }
                }
                cursor
            }
            }


        """
            % LABEL_SHARD
        )
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "cursor": cursor,
            "first": first,
        }

        result = self.client.execute_query(query_string, query_variables)

        return (
            result["tasksPaged"]["entries"],
            result["tasksPaged"]["cursor"],
        )



"""Abstract interface to upload."""

from typing import List, Dict, Optional
from pathlib import Path

import aiohttp

from redbrick.common.client import RBClient
from redbrick.common.upload import UploadControllerInterface


class UploadRepo(UploadControllerInterface):
    """Handle communication with backend relating to uploads."""

    def __init__(self, client: RBClient) -> None:
        """Construct ExportRepo."""
        self.client = client

    def create_datapoint(
        self,
        org_id: str,
        project_id: str,
        storage_id: str,
        name: str,
        items: List[str],
        labels: Optional[List[Dict]],
    ) -> str:
        """
        Create a datapoint and returns its dpId.

        Name must be unique in the project.
        """
        raise NotImplementedError()

    async def create_datapoint_async(
        self,
        aio_client: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        storage_id: str,
        name: str,
        items: List[str],
        labels: Optional[List[Dict]],
    ) -> None:
        """
        Create a datapoint and returns its dpId.

        Name must be unique in the project.
        """
        query_string = """
            mutation(
                $orgId: UUID!
                $items: [String!]!
                $name: String!
                $dsetName: String!
                $storageId: UUID!
                $lsetName: String!
                $labels: [LabelInput!]
            ) {
                createDatapoint(
                    orgId: $orgId
                    items: $items
                    name: $name
                    dsetName: $dsetName
                    storageId: $storageId
                    lsetName: $lsetName
                    labels: $labels
                ) {
                    dpId
                }
            }
        """

        query_variables = {
            "orgId": org_id,
            "items": items,
            "name": name,
            "dsetName": project_id,
            "storageId": storage_id,
            "lsetName": project_id + "-input",
            "labels": labels or [],
        }

        await self.client.execute_query_async(aio_client, query_string, query_variables)

    def upload_image(self, org_id: str, project_id: str, file_path: Path) -> str:
        """Upload a local image and add labels."""
        raise NotImplementedError()

    async def upload_image_async(
        self, org_id: str, project_id: str, file_path: Path
    ) -> str:
        """Upload a local image and add labels."""
        raise NotImplementedError()



"""Test functions for coco and rb categories."""

from .categories import rb2coco_categories_format, rb_get_class_id

taxonomy1 = {
    "name": "helloWorld",
    "version": 1,
    "attributes": [],
    "categories": [
        {
            "name": "object",
            "classId": -1,
            "children": [
                {"name": "bus", "classId": 0, "children": []},
                {"name": "traffic light", "classId": 1, "children": []},
                {"name": "traffic sign", "classId": 2, "children": []},
            ],
        }
    ],
}

taxonomy2 = {
    "name": "CocoStuff",
    "version": 1,
    "attributes": [],
    "categories": [
        {
            "name": "stuff",
            "classId": -1,
            "children": [
                {
                    "name": "outdoor",
                    "classId": 0,
                    "children": [
                        {
                            "name": "water",
                            "classId": 1,
                            "children": [
                                {"name": "water-other", "classId": 2, "children": [],},
                                {"name": "waterdrops", "classId": 3, "children": [],},
                                {"name": "sea", "classId": 4, "children": []},
                                {"name": "river", "classId": 5, "children": []},
                                {"name": "fog", "classId": 6, "children": []},
                            ],
                        },
                    ],
                }
            ],
        }
    ],
}


def test_ms_coco_categories_format() -> None:
    """Test converting into ms coco category format."""
    # arrange

    # action
    result = rb2coco_categories_format(taxonomy1)

    # asset
    assert result == [
        {"name": "bus", "id": 0, "supercategory": "object"},
        {"name": "traffic light", "id": 1, "supercategory": "object"},
        {"name": "traffic sign", "id": 2, "supercategory": "object"},
    ]


def test_ms_coco_categories_format_double_nested() -> None:
    """Test converting into ms coco category format."""
    # arrange

    # action
    result = rb2coco_categories_format(taxonomy2)

    # asset
    assert result == [
        {"name": "outdoor", "id": 0, "supercategory": "stuff"},
        {"name": "water", "id": 1, "supercategory": "outdoor"},
        {"name": "water-other", "id": 2, "supercategory": "water"},
        {"name": "waterdrops", "id": 3, "supercategory": "water"},
        {"name": "sea", "id": 4, "supercategory": "water"},
        {"name": "river", "id": 5, "supercategory": "water"},
        {"name": "fog", "id": 6, "supercategory": "water"},
    ]


def test_rb_get_class_id() -> None:
    """Test getting class id for rb category."""
    # arrange
    category = ["stuff", "outdoor", "water", "fog"]
    expected = 6

    # action
    result = rb_get_class_id(category, taxonomy2)

    # assert
    assert result == expected



"""Test getting image size."""

from redbrick.utils.image import get_image_size, url_to_image


def test_get_image_size() -> None:
    """Test getting image size."""
    # arrange
    url = "https://cdn.mos.cms.futurecdn.net/wtqqnkYDYi2ifsWZVW2MT4-1200-80.jpg"

    # action
    type_, width, height = get_image_size(url)

    # assert
    assert type_ == "jpg"
    assert width == 1000
    assert height == 667


def test_url_to_image() -> None:
    """Test getting an image from a url."""
    # arrange
    url = "https://cdn.mos.cms.futurecdn.net/wtqqnkYDYi2ifsWZVW2MT4-1200-80.jpg"

    # action
    result = url_to_image(url)

    # assert
    assert result.size



"""Abstract interface to exporting."""

from typing import Optional, List, Dict
from abc import ABC, abstractmethod
import aiohttp


class LabelingControllerInterface(ABC):
    """Abstract interface to Labeling APIs."""

    @abstractmethod
    def get_labeling_tasks(
        self, org_id: str, project_id: str, stage_name: str, count: int = 5
    ) -> List[Dict]:
        """Get labeling tasks."""

    @abstractmethod
    async def put_labeling_results(
        self,
        session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        stage_name: str,
        task_id: str,
        labels: List[Dict],
    ) -> None:
        """Put Labeling results."""

    @abstractmethod
    async def put_review_task_result(
        self,
        session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        stage_name: str,
        task_id: str,
        review_val: bool,
    ) -> None:
        """Put review result for task."""



"""Logging functions."""
from typing import Union, Optional
from termcolor import colored


def print_info(text: str, end: Optional[str] = None) -> None:
    """Log general information."""
    print(colored("[INFO]: ", "blue"), text, end=end)


def print_warning(text: str) -> None:
    """Log warnings."""
    print(colored("[WARNING]: ", "yellow"), text)


def print_error(text: Union[str, Exception]) -> None:
    """Log errors."""
    print(colored("[ERROR]: ", "red"), text)



"""Async utils."""

import asyncio
from typing import Generator, Sequence, Awaitable, List, TypeVar, Optional, Iterable
import tqdm.asyncio  # type: ignore

ReturnType = TypeVar("ReturnType")


async def gather_with_concurrency(
    max_concurrency: int,
    tasks: Iterable[Awaitable[ReturnType]],
    progress_bar_name: Optional[str] = None,
) -> List[ReturnType]:
    """Utilizes a Semaphore to limit concurrency to n."""
    semaphore = asyncio.Semaphore(max_concurrency)

    async def sem_task(task: Awaitable[ReturnType]) -> ReturnType:
        async with semaphore:
            return await task

    coros = [sem_task(task) for task in tasks]

    if progress_bar_name:
        result = []
        print(progress_bar_name)
        for f in tqdm.asyncio.tqdm.as_completed(coros):
            temp = await f
            result.append(temp)
        return result

    return await asyncio.gather(*coros)



"""Methods to handle converting category formats."""
from typing import Dict, List, Optional


def rb2coco_categories_format(taxonomy: Dict) -> List[Dict]:
    """Convert to ms_coco categories format."""
    root = taxonomy["categories"][0]

    output: List[Dict] = []

    def recurse(cat: Dict) -> None:
        """Recurse and add child entries to output."""
        for child in cat.get("children", []):

            temp = {
                "name": child["name"],
                "id": child["classId"],
                "supercategory": cat["name"],
            }

            output.append(temp)
            recurse(child)

    recurse(root)

    return output


def rb_get_class_id(label_category: List[str], taxonomy: dict) -> int:
    """Get classId for label using taxonomy."""
    categories = taxonomy["categories"]
    assert len(categories) == 1

    def find_name_in_categories(
        temp_categories: List[dict], name: str
    ) -> Optional[dict]:
        for cat in temp_categories:
            if cat["name"] == name:

                return cat
        return None

    for cat in label_category:
        new_root = find_name_in_categories(categories, cat)
        if new_root:
            categories = new_root.get("children")

    if new_root:
        class_id: int = new_root["classId"]
        return class_id
    else:
        raise Exception("Couldn't find class")



"""Enumerations for use across SDK."""

from enum import Enum


class LabelType(Enum):
    """Allowable types for labeling projects."""

    IMAGE_CLASSIFY = "IMAGE_CLASSIFY"
    IMAGE_DETECTION = "IMAGE_BBOX"
    IMAGE_KEYPOINT = "IMAGE_POINT"
    IMAGE_POLYLINE = "IMAGE_POLYLINE"
    IMAGE_POLYGON = "IMAGE_POLYGON"
    IMAGE_ELLIPSE = "IMAGE_ELLIPSE"
    IMAGE_SEGMENTATION = "IMAGE_SEGMENTATION"
    IMAGE_ALLTYPES = "IMAGE_MULTI"
    VIDEOFRAMES_DETECTION = "VIDEO_BBOX"
    VIDEOFRAMES_CLASSIFY = "VIDEO_CLASSIFY"
    VIDEOFRAMES_POLYGON = "VIDEO_POLYGON"



"""Abstract interface to Labeling APIs."""

from typing import Optional, List, Dict
import aiohttp

from redbrick.common.client import RBClient
from redbrick.common.labeling import LabelingControllerInterface
from .shards import LABEL_SHARD


class LabelingRepo(LabelingControllerInterface):
    """Implementation of manual labeling apis."""

    def __init__(self, client: RBClient) -> None:
        """Construct ExportRepo."""
        self.client = client

    def get_labeling_tasks(
        self, org_id: str, project_id: str, stage_name: str, count: int = 5
    ) -> List[Dict]:
        """Get labeling tasks."""
        query = """
        mutation assignLabelingTasks(
            $orgId: UUID!
            $projectId: UUID!
            $stageName: String!
            $count: Int!
        )  {
                assignLabelingTasks(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                count: $count
                ) {
                    orgId
                    projectId
                    stageName
                    state
                    taskId
                    taskType
                    completionTimeMs
                    progressSavedAt
                    assignedTo {
                        userId
                        loggedInUser
                    }
                    taxonomy {
                        name
                        version
                    }
                    datapoint {
                        dpId
                        itemsPresigned: items(presigned: true)
                        items(presigned: false)
                        dataType
                        name
                    }
                    taskData {
                        subName
                        taskType
                        createdAt
                        createdBy
                        labels(interpolate: true) {
                            %s
                        }
                }
            }
        }
        """ % (
            LABEL_SHARD
        )

        response = self.client.execute_query(
            query,
            {
                "orgId": org_id,
                "projectId": project_id,
                "stageName": stage_name,
                "count": count,
            },
        )
        tasks: List[Dict] = response["assignLabelingTasks"]
        return tasks

    async def put_labeling_results(
        self,
        session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        stage_name: str,
        task_id: str,
        labels: List[Dict],
    ) -> None:
        """Put Labeling results."""
        query = """
        mutation putTaskAndLabels(
        $orgId: UUID!
        $projectId: UUID!
        $stageName: String!
        $taskId: UUID!
        $elapsedTimeMs: Int!
        $labels: [LabelInput]!
        $finished: Boolean!
        ) {
            putManualLabelingTaskAndLabels(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                taskId: $taskId
                elapsedTimeMs: $elapsedTimeMs
                labels: $labels
                finished: $finished
            ) {
                ok
            }
        }
        """

        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "taskId": task_id,
            "labels": labels,
            "finished": True,
            "elapsedTimeMs": 0,
        }
        result = await self.client.execute_query_async(session, query, variables)

    async def put_review_task_result(
        self,
        session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        stage_name: str,
        task_id: str,
        review_val: bool,
    ) -> None:
        """Put review result for task."""
        query = """
        mutation putReviewTask(
        $orgId: UUID!
        $projectId: UUID!
        $stageName: String!
        $reviewVal: Boolean!
        $taskId: UUID!
        $elapsedTimeMs: Int!
        ) {
            putExpertReviewTask(
            orgId: $orgId
            projectId: $projectId
            stageName: $stageName
            reviewVal: $reviewVal
            taskId: $taskId
            elapsedTimeMs: $elapsedTimeMs
            ) {
            ok
            }
        }
        """

        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "taskId": task_id,
            "reviewVal": review_val,
            "elapsedTimeMs": 0,
        }

        result = await self.client.execute_query_async(session, query, variables)



"""Main file for converting RedBrick format to coco format."""
import asyncio
from typing import Dict, List, Tuple
from redbrick.utils.async_utils import gather_with_concurrency
from redbrick.utils import aioimgspy
from redbrick.utils.logging import print_error
import aiohttp
from yarl import URL
from .polygon import rb2coco_polygon
from .bbox import rb2coco_bbox
from .categories import rb_get_class_id, rb2coco_categories_format


async def _get_image_dimension_map(
    datapoints: List[Dict],
) -> Dict[str, Tuple[int, int]]:
    """Get a map from dpId to [width, height] of the images."""

    async def _get_size(
        session: aiohttp.ClientSession, datapoint: Dict
    ) -> Tuple[int, int, str]:
        async with session.get(
            # encode with yarl so that aiohttp doesn't encode again.
            URL(datapoint["itemsPresigned"][0], encoded=True)
        ) as response:
            temp = await aioimgspy.probe(response.content)  # type: ignore
            return (temp["width"], temp["height"], datapoint["dpId"])

    async with aiohttp.ClientSession() as session:
        coros = [_get_size(session, dpoint) for dpoint in datapoints]
        all_sizes = await gather_with_concurrency(10, coros, "Getting image dimensions")

    return {temp[2]: (temp[0], temp[1]) for temp in all_sizes}


def coco_converter(datapoints: List[Dict], taxonomy: Dict) -> Dict:
    """Convert redbrick labels to standard coco format."""
    coco_categories = rb2coco_categories_format(taxonomy)

    image_dims_map = asyncio.run(_get_image_dimension_map(datapoints))

    images: List[Dict] = []
    annotations: List[Dict] = []
    for data in datapoints:
        file_name = data["name"]
        dp_id = data["dpId"]
        name = data["name"]
        labels = data["labels"]

        width, height = image_dims_map[dp_id]

        current_image_id = len(images)
        image_entry = {
            "id": current_image_id,
            "file_name": file_name,
            "raw_url": data["items"][0],
            "signed_url": data["itemsPresigned"][0],
            "dp_id": data["dpId"],
            "height": height,
            "width": width,
        }

        for label in labels:
            annotation_index = len(annotations)
            if label.get("bbox2d"):
                class_id = rb_get_class_id(label["category"][0], taxonomy)
                coco_label = rb2coco_bbox(
                    label, annotation_index, current_image_id, class_id, width, height
                )
                annotations.append(coco_label)
            elif label.get("polygon"):
                class_id = rb_get_class_id(label["category"][0], taxonomy)
                coco_label = rb2coco_polygon(
                    label, annotation_index, current_image_id, class_id, width, height
                )
                annotations.append(coco_label)
            else:
                print_error(
                    "Label not converted, Coco format only supports bbox and polygon labels."
                )
        images.append(image_entry)

    return {
        "images": images,
        "annotations": annotations,
        "categories": coco_categories,
        "info": {},
        "licenses": [],
    }



"""
RedBrick SDK to enable powerful use cases.

- To begin use: redbrick.get_project(api_key, org_id, project_id)
- Start at https://app.redbrickai.com to create a project.
- Learn more at https://docs.redbrickai.com


"""

from redbrick.common.context import RBContext
from redbrick.common.enums import LabelType
from redbrick.project import RBProject
from redbrick.organization import RBOrganization
from redbrick.utils import version_check
from redbrick.repo import (
    ExportRepo,
    LabelingRepo,
    LearningRepo,
    UploadRepo,
    ProjectRepo,
)

version_check.version_check()


def _populate_context(context: RBContext) -> RBContext:
    context.export = ExportRepo(context.client)
    context.labeling = LabelingRepo(context.client)
    context.learning = LearningRepo(context.client)
    context.upload = UploadRepo(context.client)
    context.project = ProjectRepo(context.client)
    return context


def get_org(api_key: str, url: str, org_id: str) -> RBOrganization:
    """Get redbrick organization object."""
    context = _populate_context(RBContext(api_key, url))
    return RBOrganization(context, org_id)


def get_project(api_key: str, url: str, org_id: str, project_id: str) -> RBProject:
    """Get project object."""
    context = _populate_context(RBContext(api_key, url))
    return RBProject(context, org_id, project_id)



from .polygon import rb2coco_polygon, is_coco_polygon


def test_rb2coco_polygon_simple_square() -> None:
    """Test converting polygon rb label to coco label when the polygon is just a square."""
    # arrange
    label = {
        "label_id": "1234",
        "category": [["object", "bus"]],
        "polygon": [
            {"xnorm": 0, "ynorm": 0},
            {"xnorm": 1, "ynorm": 0},
            {"xnorm": 1, "ynorm": 1},
            {"xnorm": 0, "ynorm": 1},
        ],
    }

    # action
    result = rb2coco_polygon(label, 1, 1, 2, 100, 100)
    # result2 = coco2rb_bbox(result, [["object", "bux"]], 100, 100)
    # result3 = rb2coco_bbox(result2, 1, 1, 2, 100, 100)
    # assert
    assert result == {
        "id": 1,
        "image_id": 1,
        "category_id": 2,
        "bbox": [0, 0, 100, 100],
        "area": 10000,
        "iscrowd": 0,
        "segmentation": [[0, 0, 100, 0, 100, 100, 0, 100]],
    }
    # assert result == result3, "Conversion is not consistent/symmetric"
    assert is_coco_polygon(result)


def test_rb2coco_polygon_diamond() -> None:
    """Test converting polygon rb label to coco label when the polygon is a diamond."""
    # arrange
    label = {
        "label_id": "1234",
        "category": [["object", "bus"]],
        "polygon": [
            {"xnorm": 0.5, "ynorm": 0},
            {"xnorm": 1, "ynorm": 0.5},
            {"xnorm": 0.5, "ynorm": 1},
            {"xnorm": 0, "ynorm": 0.5},
        ],
    }

    # action
    result = rb2coco_polygon(label, 1, 1, 2, 100, 100)
    # result2 = coco2rb_bbox(result, [["object", "bux"]], 100, 100)
    # result3 = rb2coco_bbox(result2, 1, 1, 2, 100, 100)
    # assert
    assert result == {
        "id": 1,
        "image_id": 1,
        "category_id": 2,
        "bbox": [0, 0, 100, 100],
        "area": 5000,
        "iscrowd": 0,
        "segmentation": [[50, 0, 100, 50, 50, 100, 0, 50]],
    }
    # assert result == result3, "Conversion is not consistent/symmetric"
    assert is_coco_polygon(result)


def test_rb2coco_polygon_diamond_reverse_order() -> None:
    """
    Test converting polygon rb label to coco label with diamond shape.

    Reverse the direction to ensure area computation is consistent.
    """
    # arrange
    label = {
        "label_id": "1234",
        "category": [["object", "bus"]],
        "polygon": [
            {"xnorm": 0, "ynorm": 0.5},
            {"xnorm": 0.5, "ynorm": 1},
            {"xnorm": 1, "ynorm": 0.5},
            {"xnorm": 0.5, "ynorm": 0},
        ],
    }

    # action
    result = rb2coco_polygon(label, 1, 1, 2, 100, 100)

    # assert
    assert result == {
        "id": 1,
        "image_id": 1,
        "category_id": 2,
        "bbox": [0, 0, 100, 100],
        "area": 5000,
        "iscrowd": 0,
        "segmentation": [[0, 50, 50, 100, 100, 50, 50, 0]],
    }
    # assert result == result3, "Conversion is not consistent/symmetric"
    assert is_coco_polygon(result)



"""Management of versions to help users update."""

from typing import List
import json
import os

import requests
from distutils.version import StrictVersion
from .logging import print_warning
import redbrick


def version_check() -> None:
    """Check if current installed version of the SDK is up to date with latest pypi release."""
    # Getting latest version on pypi
    url = "https://pypi.org/pypi/{}/json".format("redbrick-sdk")
    data = requests.get(url).json()
    versions = list(data["releases"].keys())
    versions.sort(key=StrictVersion)
    latest_version = versions[-1]
    # Comparing with current installed version
    with open(
        os.path.join(os.path.dirname(redbrick.__file__), "VERSION"),
        "r",
        encoding="utf-8",
    ) as f:
        curr_version = f.read().strip()
    if curr_version != latest_version:
        warn = (
            "You are using version '{}' of the SDK. However, version '{}' is available!\n"
            + "Please update as soon as possible to get the latest features and bug fixes.\n"
            + "You can use 'python -m pip install --upgrade redbrick-sdk' to get the latest version."
        )
        print_warning(warn.format(curr_version, latest_version))



"""Main object for RedBrick SDK."""

import time
from typing import Dict, List
from redbrick.common.context import RBContext

from redbrick.export import Export
from redbrick.upload import Upload
from redbrick.learning import Learning
from redbrick.labeling import Labeling
from redbrick.utils.logging import print_info


class RBProject:
    """Interact with a RedBrick project."""

    def __init__(self, context: RBContext, org_id: str, project_id: str) -> None:
        """Construct RBProject."""
        self.context = context

        self._org_id = org_id
        self._project_id = project_id
        self._project_name: str
        self._stages: List[Dict]

        # check if project exists on backend to validate
        self._get_project()

        self.export = Export(context, org_id, project_id)

        self.labeling = Labeling(context, org_id, project_id)
        self.review = Labeling(context, org_id, project_id, review=True)
        self.upload = Upload(context, org_id, project_id)

    @property
    def org_id(self) -> str:
        """Read only property."""
        return self._org_id

    @property
    def project_id(self) -> str:
        """Read only property."""
        return self._project_id

    @property
    def name(self) -> str:
        """Read only name property."""
        return self._project_name

    @property
    def learning(self) -> Learning:
        """Read only, get learning module."""
        for stage in self._stages:
            if stage["brickName"] == "active-learning":
                return Learning(
                    self.context, self.org_id, self.project_id, stage["stageName"]
                )

        raise Exception("No active learning stage in this project")

    def __wait_for_project_to_finish_creating(self) -> Dict:
        project = self.context.project.get_project(self.org_id, self.project_id)
        if project["status"] == "CREATING":
            print_info("Project is still creating...", end="")
            for ii in range(8):
                print_info(".", end="")
                time.sleep(2 ^ ii)
                project = self.context.project.get_project(self.org_id, self.project_id)
                if project["status"] != "CREATING":
                    break

        if project["status"] == "REMOVING":
            raise Exception("Project has been deleted")
        if project["status"] == "CREATION_FAILURE":
            raise Exception("Project failed to be created")
        if project["status"] == "CREATION_SUCCESS":
            return project
        raise Exception("Unknown problem occurred")

    def _get_project(self) -> None:
        """Get project to confirm it exists."""
        project = self.__wait_for_project_to_finish_creating()

        self._project_name = project["name"]
        self._stages = self.context.project.get_stages(self.org_id, self.project_id)

    def __str__(self) -> str:
        """Get string representation of RBProject object."""
        return f"RedBrick Project - {self.name} - id:( {self.project_id} )"

    def __repr__(self) -> str:
        """Representation of object."""
        return str(self)



from .client import RBClient
from .export import ExportControllerInterface
from .upload import UploadControllerInterface
from .labeling import LabelingControllerInterface
from .learning import LearningControllerInterface
from .project import ProjectRepoInterface


class RBContext:
    """Basic context for accessing low level functionality."""

    def __init__(self, api_key: str, url: str) -> None:
        """Construct RedBrick client singleton."""
        self.client = RBClient(api_key, url)

        self.export: ExportControllerInterface
        self.upload: UploadControllerInterface
        self.labeling: LabelingControllerInterface
        self.learning: LearningControllerInterface
        self.project: ProjectRepoInterface

    def __str__(self) -> str:
        """Get string representation."""
        return repr(self) + "***" + self.client.api_key[-4:-1]



"""Utils for getting image size."""

from typing import Tuple
import io

import imgspy  # type: ignore
from PIL import Image  # type: ignore

import requests


def get_image_size(url: str) -> Tuple[str, int, int]:
    """Attempt to get size of image from URL."""
    with requests.get(url, stream=True) as res:
        result = imgspy.info(res.raw)

    return result["type"], result["width"], result["height"]


def url_to_image(url: str) -> Image:
    """Get image data from url."""
    data = requests.get(url).content
    return Image.open(io.BytesIO(data))



"""Public interface to upload module."""

import asyncio
from copy import deepcopy
from typing import List, Dict, Optional

import aiohttp

from redbrick.common.context import RBContext
from redbrick.utils.async_utils import gather_with_concurrency


class Upload:
    """Primary interface to uploading new data to a project."""

    def __init__(self, context: RBContext, org_id: str, project_id: str) -> None:
        """Construct Upload object."""
        self.context = context
        self.org_id = org_id
        self.project_id = project_id

    async def _create_datapoint(
        self, session: aiohttp.ClientSession, storage_id: str, point: Dict
    ) -> Optional[Dict]:
        """Try to create a datapoint."""
        try:
            await self.context.upload.create_datapoint_async(
                session,
                self.org_id,
                self.project_id,
                storage_id,
                point["name"],
                point["items"],
                point.get("labels"),
            )
        except Exception as error:
            print(error)
            point_error = deepcopy(point)
            point_error["error"] = error
            return point_error
        return None

    async def _create_datapoints(
        self, storage_id: str, points: List[Dict]
    ) -> List[Dict]:
        async with aiohttp.ClientSession() as session:
            coros = [
                self._create_datapoint(session, storage_id, point) for point in points
            ]

            temp = await gather_with_concurrency(50, coros, "Creating datapoints")
            failed = []
            for val in temp:
                if val:
                    failed.append(val)
            return failed

    def create_datapoints(self, storage_id: str, points: List[Dict]) -> List[Dict]:
        """
        Create datapoints in project.

        Returns list of datapoints that failed to create.
        """
        return asyncio.run(self._create_datapoints(storage_id, points))



"""Convert between rb and coco for polygon labels."""

from typing import Dict, List
from shapely.geometry import Polygon  # type: ignore


def _create_bbox(polygon: Polygon) -> List[int]:
    """Get a bounding box entry in a single COCO annotation field for a polygon."""
    bounds = polygon.bounds

    x_min = bounds[0]
    y_min = bounds[1]
    x_max = bounds[2]
    y_max = bounds[3]

    width = x_max - x_min
    height = y_max - y_min

    return [x_min, y_min, width, height]


def _get_polygon(segmentation: List[int]) -> Polygon:
    """Get shapely polygon object."""
    x_pos = []
    y_pos = []

    for idx, coord in enumerate(segmentation):
        if idx % 2 == 0:
            x_pos += [coord]
        else:
            y_pos += [coord]

    return Polygon(zip(x_pos, y_pos))


def _get_segmentation(polygon: List[Dict], width: int, height: int) -> List[int]:
    """Convert RedBrick polygon to coco segmentation field."""
    segment = []
    for point in polygon:
        segment += [
            int(point["xnorm"] * width),
            int(point["ynorm"] * height),
        ]

    return segment


def rb2coco_polygon(
    rb_label: Dict,
    label_id: int,
    image_id: int,
    category_id: int,
    width: int,
    height: int,
) -> Dict:
    """Convert rb polygon to coco polygon."""
    assert rb_label.get("polygon")

    segmentation = _get_segmentation(rb_label["polygon"], width, height)
    polygon = _get_polygon(segmentation)
    bbox = _create_bbox(polygon)

    return {
        "id": label_id,
        "image_id": image_id,
        "category_id": category_id,
        "bbox": bbox,
        "iscrowd": 0,
        "area": int(polygon.area),
        "segmentation": [segmentation],
    }


def is_coco_polygon(coco_label: Dict) -> bool:
    """Check if a given label is a valid polygon only label."""
    has_bbox = "bbox" in coco_label and len(coco_label["bbox"]) == 4
    has_segment = len(coco_label.get("segmentation", [])) > 0
    no_keypoints = not coco_label.get("keypoints")
    return has_bbox and has_segment and no_keypoints


def coco2rb_polygon(
    coco_label: Dict, category: List[List[str]], width: int, height: int
) -> Dict:
    """Convert coco polygon to rb polygon."""



"""Setup the python module."""

from setuptools import setup, find_packages  # type: ignore
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

with open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "redbrick", "VERSION"),
    "r",
    encoding="utf-8",
) as f:
    version = f.read().strip()

setup(
    name="redbrick-sdk",
    url="https://github.com/redbrick-ai/redbrick-sdk",
    version=version,
    description="RedBrick platform Python SDK!",
    py_modules=["redbrick"],
    packages=find_packages(),
    # package_dir={"": "redbrick"},
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=[
        "requests==2.23.0",
        "tqdm==4.50.0",
        "termcolor==1.1.0",
        "requests==2.23.0",
        "aiohttp==3.7.4",
        "cchardet==2.1.7",
        "aiodns==3.0.0",
        "Pillow==8.2.0",
        "shapely==1.7.1",
    ],
    extras_require={
        "dev": [
            "pytest>=3.7",
            "black==19.10b0",
            "pydocstyle==5.0.2",
            "pycodestyle==2.6.0",
            "twine==3.1.1",
            "wheel==0.35.1",
            "mypy==0.790",
            "pytest-mock==3.3.1",
        ]
    },
)



"""Public interface to labeling module."""

import asyncio
from typing import List, Dict, Optional
from copy import deepcopy
import aiohttp

from redbrick.common.context import RBContext
from redbrick.utils.async_utils import gather_with_concurrency
from redbrick.utils.rb_label_utils import clean_rb_label


class Labeling:
    """Perform automated labeling tasks."""

    def __init__(
        self, context: RBContext, org_id: str, project_id: str, review: bool = False
    ) -> None:
        """Construct Labeling."""
        self.context = context
        self.org_id = org_id
        self.project_id = project_id
        self.review = review

    def get_tasks(self, stage_name: str, count: int = 1) -> List[Dict]:
        """Get a list of tasks."""
        tasks = self.context.labeling.get_labeling_tasks(
            self.org_id, self.project_id, stage_name, count=count
        )

        def _clean_tasks(task: Dict) -> Dict:
            task_id = task["taskId"]
            labels = task.get("taskData", {}).get("labels", [])
            labels_cleaned = [clean_rb_label(label) for label in labels]

            items = task["datapoint"]["items"]
            items_presigned = task["datapoint"]["itemsPresigned"]
            name = task["datapoint"]["name"]

            return {
                "taskId": task_id,
                "labels": labels,
                "items": items,
                "itemsPresigned": items_presigned,
                "name": name,
            }

        return [_clean_tasks(task) for task in tasks]

    async def _put_task(
        self, session: aiohttp.ClientSession, stage_name: str, task: Dict
    ) -> Optional[Dict]:
        task_id = task["taskId"]
        try:
            if self.review:
                review_val = task["reviewVal"]
                await self.context.labeling.put_review_task_result(
                    session,
                    self.org_id,
                    self.project_id,
                    stage_name,
                    task_id,
                    review_val,
                )
            else:

                labels = task["labels"]
                await self.context.labeling.put_labeling_results(
                    session, self.org_id, self.project_id, stage_name, task_id, labels
                )

        except Exception as error:
            print(error)
            point_error = deepcopy(task)
            point_error["error"] = error
            return point_error
        return None

    async def _put_tasks(self, stage_name: str, tasks: List[Dict]) -> List[Dict]:
        async with aiohttp.ClientSession() as session:
            coros = [self._put_task(session, stage_name, task) for task in tasks]

            temp = await gather_with_concurrency(10, coros, "Uploading tasks")
            failed = []
            for val in temp:
                if val:
                    failed.append(val)
            return failed

    def put_tasks(self, stage_name: str, tasks: List[Dict]) -> List[Dict]:
        """Put tasks, return tasks that failed."""
        return asyncio.run(self._put_tasks(stage_name, tasks))



"""Organization class."""

from typing import List
from tqdm import tqdm  # type: ignore

from redbrick.common.enums import LabelType
from redbrick.common.context import RBContext
from redbrick.project import RBProject
from .basic_project import get_basic_project


class RBOrganization:
    """Representation of RedBrick organization."""

    def __init__(self, context: RBContext, org_id: str) -> None:
        """Construct RBOrganization."""
        self.context = context

        self._org_id = org_id
        self._name: str

        self._get_org()

    def _get_org(self) -> None:
        org = self.context.project.get_org(self._org_id)
        self._name = org["name"]

    def taxonomies(self) -> List[str]:
        """Get a list of taxonomy names in the organization."""
        return self.context.project.get_taxonomies(self._org_id)

    def projects(self) -> List[RBProject]:
        """Get a list of active projects in the organization."""
        projects = self.context.project.get_projects(self._org_id)
        projects = list(filter(lambda x: x["status"] == "CREATION_SUCCESS", projects))
        return [
            RBProject(self.context, self._org_id, proj["projectId"])
            for proj in tqdm(projects)
        ]

    @property
    def org_id(self) -> str:
        """Get org_id read only field."""
        return self._org_id

    @property
    def name(self) -> str:
        """Get name of organization."""
        return self._name

    def __str__(self) -> str:
        """Get string representation of RBOrganization object."""
        return f"RedBrick AI Organization - {self._name} - ( {self._org_id} )"

    def __repr__(self) -> str:
        """Representation of object."""
        return str(self)

    def create_project(
        self, name: str, label_type: LabelType, taxonomy_name: str, reviews: int = 0
    ) -> RBProject:
        """Create a project, similar to quickstart through the UI."""
        stages = get_basic_project(reviews)

        project_data = self.context.project.create_project(
            self.org_id, name, stages, label_type.value, taxonomy_name
        )

        return RBProject(self.context, self.org_id, project_data["projectId"])



"""Handlers to access APIs for getting projects."""
import json
from typing import List, Dict

from redbrick.common.client import RBClient
from redbrick.common.project import ProjectRepoInterface


class ProjectRepo(ProjectRepoInterface):
    """Class to manage interaction with project APIs."""

    def __init__(self, client: RBClient) -> None:
        """Construct ProjectRepo."""
        self.client = client

    def get_project(self, org_id: str, project_id: str) -> Dict:
        """
        Get project name and status.

        Raise an exception if project does not exist.
        """
        query = """
            query sdkGetProjectName($orgId: UUID!, $projectId: UUID!){
                project(orgId: $orgId, projectId: $projectId){
                    orgId
                    projectId
                    name
                    status
                }
            }
        """
        variables = {"orgId": org_id, "projectId": project_id}
        response: Dict[str, Dict] = self.client.execute_query(query, variables)
        if response.get("project"):
            return response["project"]

        raise Exception("Project does not exist")

    def get_stages(self, org_id: str, project_id: str) -> List[Dict]:
        """Get stages."""
        query = """
            query sdkGetStages($orgId: UUID!, $projectId: UUID!){
                stages(orgId: $orgId, projectId: $projectId){
                    stageName
                    brickName
                }
            }
        """
        variables = {"orgId": org_id, "projectId": project_id}
        response: Dict[str, List[Dict]] = self.client.execute_query(query, variables)
        return response["stages"]

    def create_project(
        self, org_id: str, name: str, stages: List[dict], td_type: str, tax_name: str
    ) -> Dict:
        """Create a project and return project_id."""
        query = """
            mutation createProjectSimple(
                $orgId: UUID!
                $name: String!
                $stages: [StageInputSimple!]!
                $tdType: TaskDataType!
                $taxonomyName: String!
                $taxonomyVersion: Int!
            ) {
                createProjectSimple(
                orgId: $orgId
                name: $name
                stages: $stages
                tdType: $tdType
                taxonomyName: $taxonomyName
                taxonomyVersion: $taxonomyVersion
                ) {
                ok
                errors
                project {
                    projectId
                    name
                    desc
                }
                stages {
                    stageName
                    brickName
                }
                }
            }
        """
        for stage in stages:
            stage["stageConfig"] = json.dumps(stage["stageConfig"])
        variables = {
            "orgId": org_id,
            "name": name,
            "stages": stages,
            "tdType": td_type,
            "taxonomyName": tax_name,
            "taxonomyVersion": 1,
        }

        response: Dict[str, Dict] = self.client.execute_query(query, variables)
        return {
            "orgId": org_id,
            "projectId": response["createProjectSimple"]["project"]["projectId"],
        }

    def get_org(self, org_id: str) -> Dict:
        """Get organization."""
        query = """
            query getOrgSDK($orgId: UUID!) {
                organization(orgId: $orgId){
                    name
                    orgId
                }
            }
        """
        response: Dict[str, Dict] = self.client.execute_query(query, {"orgId": org_id})
        return response["organization"]

    def get_projects(self, org_id: str) -> List[Dict]:
        """Get all projects in organization."""
        query = """
            query getProjectsSDK($orgId: UUID!) {
                projects(orgId: $orgId) {
                    orgId
                    name
                    projectId
                    status
                    desc
                }
            }
        """
        response: Dict[str, List[Dict]] = self.client.execute_query(
            query, {"orgId": org_id}
        )
        return response["projects"]

    def get_taxonomies(self, org_id: str) -> List[str]:
        """Get a list of taxonomies."""
        query = """
            query getTaxonomiesSDK($orgId: UUID!) {
                taxonomies(orgId: $orgId) {
                    orgId
                    name
                }
            }
        """
        response: Dict[str, List[Dict[str, str]]] = self.client.execute_query(
            query, {"orgId": org_id}
        )
        return [tax["name"] for tax in response["taxonomies"]]



"""Abstract interface to exporting data from a project."""

from typing import Optional, List, Dict, Tuple
from abc import ABC, abstractmethod


class ExportControllerInterface(ABC):
    @abstractmethod
    def get_datapoints_input(
        self,
        org_id: str,
        project_id: str,
        first: int = 50,
        presign: bool = False,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get datapoints that were uploaded to the project."""

    @abstractmethod
    def get_output_info(self, org_id: str, project_id: str,) -> Dict:
        """Get info about the output labelset and taxonomy."""

    @abstractmethod
    def get_datapoints_output(
        self,
        org_id: str,
        project_id: str,
        first: int = 50,
        presign: bool = False,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get datapoints that have made it to the output of the project."""

    @abstractmethod
    def datapoints_in_project(self, org_id: str, project_id: str) -> int:
        """Get number of datapoints in project."""

    @abstractmethod
    def get_datapoints_latest(
        self,
        org_id: str,
        project_id: str,
        first: int = 50,
        presign: bool = False,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get the latest datapoints."""



"""Abstract interface to upload."""

from typing import List, Dict, Optional
from abc import ABC, abstractmethod

from pathlib import Path

import aiohttp


class UploadControllerInterface(ABC):
    @abstractmethod
    def create_datapoint(
        self,
        org_id: str,
        project_id: str,
        storage_id: str,
        name: str,
        items: List[str],
        labels: Optional[List[Dict]],
    ) -> str:
        """
        Create a datapoint and returns its dpId.

        Name must be unique in the project.
        """

    @abstractmethod
    async def create_datapoint_async(
        self,
        aio_http_session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        storage_id: str,
        name: str,
        items: List[str],
        labels: Optional[List[Dict]],
    ) -> None:
        """
        Create a datapoint and returns its dpId.

        Name must be unique in the project.
        """

    @abstractmethod
    def upload_image(self, org_id: str, project_id: str, file_path: Path,) -> str:
        """Upload a local image and add labels."""

    @abstractmethod
    async def upload_image_async(
        self, org_id: str, project_id: str, file_path: Path,
    ) -> str:
        """Upload a local image and add labels."""
